#include "ttf.h"
#include "font.h"
#include "loadtabl.h"
#include "ttf2pk.h"
#include "ttf2bmp.h"


void PSCurveto(OUTLINE *ol,char *buf,int x, int y, int s, int t);

/*
   render a solid font.
*/

void Font_Render_Char(Font *f,BITMAP *bitmap,BYTE *cc)
{
  int index,i;
  OUTLINE *ol;
  USHORT EM,totpoint;

  /* 
     get the start address of outline data. The Font_select_CMAP()
     should be called first to setup character-to-index table.
  */
  index = Font_getFontOffset(f,cc);
  if (index == -1)
  {
    memset(BITMAP_map(bitmap),0,BITMAP_len(bitmap));
    return;
  }
  /* load outline data to f->glyf */
  Font_load_GLYF(f,f->fp,index,0);
  /* 
     if we have a data error or the index has no data or the index points to
     a composite font.
  */
  if (f->glyf->valid==0)
  {
    /* fill bitmap with blanks */
    memset(bitmap->map,0,bitmap->xsize*bitmap->ysize);
    return;
  }
  ol = f->glyf;
  EM = f->head->unitsPerEm;
  totpoint= ol->endPtsOfContours[ol->numberOfContours-1]+1;
  /* rescale the outline data to the size described by (dpoint,res) */
  for(i=0;i<totpoint;i++)
  {
    ol->xCoordinates[i] = 
      (ol->xCoordinates[i] - f->head->xMin)*bitmap->xpoint/(EM/4)-2;
    ol->yCoordinates[i] = 
      (ol->yCoordinates[i] - f->head->yMin)*bitmap->ysize/(EM/4)-2;
  }
  /* 
     call a block of stolen code :-) to convert outline data
     into bitmap data. This block deserves rewriting.
  */
  convert(f,bitmap);
}

void PSMoveto(int x,int y,char *buf)
{
  sprintf(buf,"%d %d moveto\n",x,y) ;
}

void PSLineto(int x,int y,char *buf)
{
  sprintf(buf,"%d %d lineto\n",x,y) ;
}

#define xcoor ol->xCoordinates
#define ycoor ol->yCoordinates      

void Font_Render_Char_PS(Font *f,BYTE *cc,char *buf,int len)
{
  int i,j;
  int index;
  OUTLINE *ol;
  USHORT totpoint;
  int fst,start_offpt,end_offpt=0;
  /* 
     get the start address of outline data. The Font_select_CMAP()
     should be called first to setup character-to-index table.
  */
  index = Font_getFontOffset(f,cc);

  /* load outline data to f->glyf */
  Font_load_GLYF(f,f->fp,index,0);
  /* 
     if we have a data error or the index has no data or the index points to
     a composite font.
  */
  if (f->glyf->valid==0)
  {
    /* fill bitmap with blink */
    *buf = 0;
    return;
  }
  ol = f->glyf;
  totpoint= ol->endPtsOfContours[ol->numberOfContours-1]+1;
  for (i = 0, j = 0; i < ol->numberOfContours; i++) 
  { 
    fst = j; 
    PSMoveto(xcoor[j],ycoor[j],buf); 
    buf += strlen(buf);
    start_offpt = 0; /*start at least 1*/ 

    /* data pts for all contours stored in one array. 
       Each round j init at last j + 1 */ 

    /* start_offpt means start of off points.  
       0 means no off points in record. 
       N means the position of the off point. 
       end_offpt means the ending off point. 
       (lastx, lasty) is the last ON point from which Curve and Line  
       shall start. 
    */ 

    /* start with j=0. into loop, j=1.  
       if pt[1] off, if start_offpt == 0, toggle start_offpt 
       next j=2. if on, now start_off != 0, run Curveto. 
       if pt[1] on, start_off == 0, will run Lineto. 
    */       
    for (j++; j <= ol->endPtsOfContours[i]; j++) 
    { 
      if (!(ol->flags[j]&1)) 
      { /*Off curve*/ 
        if (!start_offpt) { start_offpt = end_offpt = j; } 
        else end_offpt++; 
      } 
      else
      { /*On Curve*/ 
        if (start_offpt) 
        { 
          /* start_offpt stuck at j, end_offpt++.  
             end_offpt - start_offpt gives no of off 
             pts. start_offpt gives start of sequence. 
             why need xcoor[j] ycoor[j]? 
          */ 
          PSCurveto(ol,buf,xcoor[j],ycoor[j],start_offpt,end_offpt); 
          buf += strlen(buf);
          start_offpt = 0; 

          /* also use start_offpt as indicator to
             save one variable!! */ 
          /* after curveto, reset condition. */ 

        } 
        else 
        {
          PSLineto(xcoor[j], ycoor[j],buf); 
          buf += strlen(buf);
        }
        /* uses lastx, lasty */ 
        /*Lineto(xcoor[j],ycoor[j],buf);*/ 
      } 
    } 
    /* looks like closepath  
    fst = first, i.e. go back to first */ 
    if (start_offpt) 
    {
      PSCurveto(ol,buf,xcoor[fst],ycoor[fst],start_offpt,end_offpt); 
      buf += strlen(buf);
    }
    else 
    {
      PSLineto(xcoor[fst],ycoor[fst],buf); 
      buf += strlen(buf);
    }
  } 
}

void PSCurveto(OUTLINE *ol,char *buf,int x, int y, int s, int t)
{ 
  int N, i; 
  double sx[3], sy[3], cx[4], cy[4]; 

  N = t-s+2; 
  for(i=0; i<N-1; i++)
  {
    sx[0] = i==0?xcoor[s-1]:(xcoor[i+s]+xcoor[i+s-1])/2; 
    sy[0] = i==0?ycoor[s-1]:(ycoor[i+s]+ycoor[i+s-1])/2; 
    sx[1] = xcoor[s+i]; 
    sy[1] = ycoor[s+i]; 
    sx[2] = i==N-2?x:(xcoor[s+i]+xcoor[s+i+1])/2; 
    sy[2] = i==N-2?y:(ycoor[s+i]+ycoor[s+i+1])/2; 
    cx[3] = sx[2]; 
    cy[3] = sy[2]; 
    cx[1] = (2*sx[1]+sx[0])/3; 
    cy[1] = (2*sy[1]+sy[0])/3; 
    cx[2] = (sx[2]+2*sx[1])/3; 
    cy[2] = (sy[2]+2*sy[1])/3; 
    sprintf(buf,"%g %g %g %g %g %g curveto\n", 
            cx[1],cy[1],cx[2],cy[2],cx[3],cy[3]); 
  } 
} 
