% This is the cweb file mule2cjk.w of the CJK Package Ver. 2.5  10-Apr-1995
%
% To print this CWEB file you should (but not must) use the CWEAVE of the
% c2cweb-package (found at the CTAN archives, e.g. pip.shsu.edu) and then say
%
%           cweave +a mule2cjk.w
%
% This (fully compatible) CWEAVE can transform CWEB-files with alternative
% output rules (look at the position of braces below!) the author (it's me
% too :-) prefer. Otherwise this file will be formatted traditionally.

\def\title{mule2cjk (CJK Version 2.5)}

\def\topofcontents{
  \null\vfill
  \centerline{\titlefont The {\ttitlefont mule2cjk} program}
  \vskip 20pt
  \centerline{(CJK Version 2.5)}
  \vfill}

\def\botofcontents{
  \vfill
  \noindent
  Copyright \copyright\ 1995 by Werner Lemberg
  \bigskip\noindent
  Permission is granted to make and distribute verbatim copies of this
  document provided that the copyright notice and this permission notice
  are preserved on all copies.

  \smallskip\noindent
  Permission is granted to copy and distribute modified versions of this
  document under the conditions for verbatim copying, provided that the
  entire resulting derived work is distributed under the terms of a
  permission notice identical to this one.}

\pageno=\contentspagenumber \advance\pageno by 1
\let\maybe=\iftrue
\fullpageheight=240mm
\pageheight=223mm
\pagewidth=158mm
\setpage
\frenchspacing

\noinx
\nosecs
\nocon

@* Function and Use. 
This small program will convert text written with multilingual Emacs (MULE,
one of the most powerful editors available for Unix) into a form which can
be processed with the \.{CJK} package (\.{CJK} enables La\TeX~2e to output
Chinese/Japanese/Korean in various encodings). \.{mule2cjk} understands only
the internal encoding of MULE, so you must either write into your \.{.emacs}
file

\hskip 2em \.{(set-default-file-coding-system *internal*)}

to have internal encoding as the default or convert your text to internal
encoding with the \.{coco} program which is part of the MULE distribution.

Use this program as a filter:

\hskip 2em \.{mule2cjk [-switches] < input\_file > output\_file}


@* The program.
In internal MULE encoding, each non-ASCII encoded character has an
additional byte in front of the character to indicate the encoding. As an
example, Latin-1 encoded characters have the byte |0x81| as a prefix, two
byte encoded Japanese characters in JIS encoding have |0x92| as a prefix
and so on.

All prefix characters which represent a new valid encoding of the \.{CJK}
package will be translated into the appropriate \.{\\CJKenc} macro; the
character which follows will be then output. Big~5 encoded characters get a
special treatment by MULE and by \.{mule2cjk} because they need two prefix
bytes (|0x98| and |0x99|) in MULE to cover the whole code range, and
additionally Big~5 characters can contain the characters `\.{\\}', `\.{\{}',
and `\.{\}}' as second bytes (which will be treated specially by \TeX), thus
we always signal for \.{CJK} a so-called preprocessed file by defining a
certain \TeX\ macro at the very beginning of the output and represent each
Big~5 encoded character in the same way as \.{Bg5conv}, the Big~5
preprocessor of the \.{CJK} package would do.

The various Latin encodings will be translated directly into \TeX\ commands
by means of lookup tables. If you activate the switch \.{-g}, German umlauts
coming from the Latin-1 encoding will be converted into macros for
\.{german.sty}; this is because German umlaut dots will be set lower with
\.{german.sty} than \TeX\ would do.

The following code is very simple. No error detection is done because \TeX\
which will see the output of \.{mule2cjk} complains loudly if something is
wrong.

@
In the following sections you will find the lookup tables for Latin
encodings.

@c
char latin_1[96][15] = {@/
    "~", "!`", "", "\\pounds",                          /* |0xA0| */@/
    "", "", "", "\\S",@/
    "\\\"{}", "\\copyright", "${}^{\\b{a}}$", "$\\ll$",@/
    "\\neg", "\\-", "", "\\={}",@/

    "${}^\\circ$", "\\pm", "${}^2$", "${}^3$",          /* |0xB0| */@/
    "\\\'{}", "$\\mu$", "\\P", "$\\cdot$",@/
    "\\c{}", "${}^1$", "${}^{\\b{o}}$", "$\\gg$",@/
    "$\\frac{1}{4}$", "$\\frac{1}{2}$", "$\\frac{3}{4}$", "?`",@/

    "\\`A", "\\\'A", "\\^A", "\\~A",                    /* |0xC0| */@/
    "\\\"A", "{\\AA}", "{\\AE}", "\\c{C}",@/
    "\\`E", "\\\'E", "\\^E", "\\\"E",@/
    "\\`I", "\\\'I", "\\^I", "\\\"I",@/
                       
    "", "\\~N", "\\`O", "\\\'O",                        /* |0xD0| */@/
    "\\^O", "\\~O", "\\\"O", "$\\times$",@/
    "{\\O}", "\\`U", "\\\'U", "\\^U",@/
    "\\\"U", "\\\'Y", "", "{\\ss}",@/
                       
    "\\`a", "\\\'a", "\\^a", "\\~a",                    /* |0xE0| */@/
    "\\\"a", "{\\aa}", "{\\ae}", "\\c{c}",@/
    "\\`e", "\\\'e", "\\^e", "\\\"e",@/
    "\\`i", "\\\'i", "\\^i", "\\\"i",@/
                       
    "", "\\~n", "\\`o", "\\\'o",                        /* |0xF0| */@/
    "\\^o", "\\~o", "\\\"o", "$\\div$",@/
    "{\\o}", "\\`u", "\\\'u", "\\^u",@/
    "\\\"u", "\\\'y", "", "\\\"y"};

@
@c
char latin_2[96][15] = {@/
    "~", "", "\\u{}", "{\\L}",                          /* |0xA0| */@/
    "", "\\v{L}", "\\\'S", "\\S",@/
    "\\\"{}", "\\v{S}", "\\c{S}", "\\v{T}",@/
    "\\\'Z", "\\-", "\\v{Z}", "\\.Z",@/

    "${}^\\circ$", "", "", "{\\l}",                     /* |0xB0| */@/
    "\\\'{}", "\\v{l}", "\\\'s", "\\v{}",@/
    "\\c{}", "\\v{s}", "\\c{s}", "\\v{t}",@/
    "\\\'z", "\\H{}", "\\v{z}", "\\.z",@/

    "\\\'R", "\\\'A", "\\^A", "\\u{A}"                  /* |0xC0| */@/
    "\\\"A", "\\\'L", "\\\'C", "\\c{C}",@/
    "\\v{C}", "\\\'E", "", "\\\"E",@/
    "\\v{E}", "\\\'I", "\\^I", "\\v{D}",@/
                       
    "", "\\\'N", "\\v{N}", "\\\'O",                     /* |0xD0| */@/
    "\\^O", "\\H{O}", "\\\"O", "$\\times$",@/
    "\\v{R}", "", "\\\'U", "\\H{U}",@/
    "\\\"U", "\\\'Y", "\\c{T}", "{\\ss}",@/
                       
    "\\\'r", "\\\'a", "\\^a", "\\u{a}",                 /* |0xE0| */@/
    "\\\"a", "\\\'l", "\\\'c", "\\c{c}",@/
    "\\v{c}", "\\\'e", "", "\\\"e",@/
    "\\v{e}", "\\\'i", "\\^i", "\\v{d}",@/
                       
    "", "\\\'n", "\\v{n}", "\\\'o",                     /* |0xF0| */@/
    "\\^o", "\\H{o}", "\\\"o", "$\\div$",@/
    "\\v{r}", "", "\\\'u", "\\H{u}",@/
    "\\\"u", "\\\'y", "\\c{t}", "\\.{}"};

@
@c
char latin_3[96][15] = {@/
    "~", "", "\\u{}", "\\pounds",                       /* |0xA0| */@/
    "", "", "\\^H", "\\S",@/
    "\\\"{}", "\\.I", "\\c{S}", "\\u{G}",@/
    "\\^J", "\\-", "", "\\.Z",@/

    "${}^\\circ$", "", "${}^2$", "${}^3$",              /* |0xB0| */@/
    "\\\'{}", "$\\mu$", "\\^h", "$\\cdot$",@/
    "\\c{}", "\\i", "\\c{s}", "\\u{g}",@/
    "\\^{\\j}", "$\\frac{1}{2}$", "$\\frac{3}{4}$", "\\.z",@/

    "\\`A", "\\\'A", "\\^A", "\\~A",                    /* |0xC0| */@/
    "\\\"A", "\\.C", "\\^C", "\\c{C}",@/
    "\\`E", "\\\'E", "\\^E", "\\\"E",@/
    "\\`I", "\\\'I", "\\^I", "\\\"I",@/
                       
    "", "\\~N", "\\`O", "\\\'O",                        /* |0xD0| */@/
    "\\^O", "\\.G", "\\\"O", "$\\times$",@/
    "\\^G", "\\`U", "\\\'U", "\\^U",@/
    "\\\"U", "\\u{U}", "\\^S", "{\\ss}",@/
                       
    "\\`a", "\\\'a", "\\^a", "\\~a",                    /* |0xE0| */@/
    "\\\"a", "\\.c", "\\^c", "\\c{c}",@/
    "\\`e", "\\\'e", "\\^e", "\\\"e",@/
    "\\`i", "\\\'i", "\\^i", "\\\"i",@/
                       
    "", "\\~n", "\\`o", "\\\'o",                        /* |0xF0| */@/
    "\\^o", "\\.g", "\\\"o", "$\\div$",@/
    "\\^g", "\\`u", "\\\'u", "\\^u",@/
    "\\\"u", "\\u{u}", "\\^s", "\\.{}"};

@
@c
char latin_4[96][15] = {@/
    "~", "", "", "\\c{R}",                              /* |0xA0| */@/
    "", "\\~I", "\\c{L}", "\\S",@/
    "\\\"{}", "\\v{S}", "\\=E", "\\c{G}",@/
    "", "\\-", "\\v{Z}", "\\={}",@/

    "${}^\\circ$", "", "", "\\c{r}",                    /* |0xB0| */@/
    "\\\'{}", "\\~{\\i}", "\\c{l}", "\\v{}",@/
    "\\c{}", "\\v{s}", "\\=e", "\\c{g}",@/
    "", "", "\\v{z}", "",@/

    "\\=A", "\\\'A", "\\^A", "\\~A",                    /* |0xC0| */@/
    "\\\"A", "{\\AA}", "{\\AE}", "",@/
    "\\v{C}", "\\\'E", "", "\\\"E",@/
    "\\.E", "\\\'I", "\\^I", "\\=I",@/
                       
    "", "\\c{N}", "\\=O", "\\c{K}",                     /* |0xD0| */@/
    "\\^O", "\\~O", "\\\"O", "$\\times$",@/
    "{\\O}", "", "\\\'U", "\\^U",@/
    "\\\"U", "\\~U", "\\=U", "{\\ss}",@/
                       
    "\\=a", "\\\'a", "\\^a", "\\~a",                    /* |0xE0| */@/
    "\\\"a", "{\\aa}", "{\\ae}", "",@/
    "\\v{c}", "\\\'e", "", "\\\"e",@/
    "\\.e", "\\\'i", "\\^i", "\\={\\i}",@/
                       
    "", "\\c{n}", "\\=o", "\\c{k}",                     /* |0xF0| */@/
    "\\^o", "\\~o", "\\\"o", "$\\div$",@/
    "{\\o}", "", "\\\'u", "\\^u",@/
    "\\\"u", "\\~u", "\\=u", "\\.{}"};

@
As you might have guessed, |main()| consists mainly of a great switch with
branches for each prefix byte.

@c
#include <getopt.h>
#include <stdio.h>@#


int main(argc, argv)
  int argc;
  char *argv[];
   {int ch, ch1, ch2;
    int out1, out2;
    int temp;
    int last_enc = 0;


    while((temp = getopt(argc, argv, "g")) != EOF)
       {switch(temp)
           {case 'g':
                strcpy(latin_1[0xAD - 0xA0], "\"-");
                strcpy(latin_1[0xC4 - 0xA0], "\"A");
                strcpy(latin_1[0xD6 - 0xA0], "\"O");
                strcpy(latin_1[0xDC - 0xA0], "\"U");
                strcpy(latin_1[0xDF - 0xA0], "\"s");
                strcpy(latin_1[0xE4 - 0xA0], "\"a");
                strcpy(latin_1[0xF6 - 0xA0], "\"o");
                strcpy(latin_1[0xFC - 0xA0], "\"u");
                break;
           }
       }@#

    fprintf(stdout, "\\def\\CJKpreproc{}\n");@#

    ch = fgetc(stdin);@#

    while(!feof(stdin))
       {switch(ch)
           {case 129: /* Latin-1 */
                ch1 = fgetc(stdin);
                if(ch1 >= 0xA0)
                    fprintf(stdout, "%s", latin_1[ch1 - 0xA0]);
                break;
            case 130: /* Latin-2 */
                ch1 = fgetc(stdin);
                if(ch1 >= 0xA0)
                    fprintf(stdout, "%s", latin_2[ch1 - 0xA0]);
                break;
            case 131: /* Latin-3 */
                ch1 = fgetc(stdin);
                if(ch1 >= 0xA0)
                    fprintf(stdout, "%s", latin_3[ch1 - 0xA0]);
                break;
            case 132: /* Latin-4 */
                ch1 = fgetc(stdin);
                if(ch1 >= 0xA0)
                    fprintf(stdout, "%s", latin_4[ch1 - 0xA0]);
                break;
            case 141: /* Latin-5 */
                break; /* not implemented yet */@#
                
            case 137: /* Japanese Katakana */
                break; /* not implemented yet */
            case 138: /* Japanese Roman */
                break; /* not implemented yet */
            case 146: /* Japanese JIS */
                if(last_enc != 146)
                    fprintf(stdout, "\\CJKenc{JIS}");
                ch1 = fgetc(stdin);
                ch2 = fgetc(stdin);
                fputc(ch1, stdout);
                fputc(ch2, stdout);
                break;@#
                
            case 145: /* Chinese GB */
                if(last_enc != 145)
                    fprintf(stdout, "\\CJKenc{GB}");
                ch1 = fgetc(stdin);
                ch2 = fgetc(stdin);
                fputc(ch1, stdout);
                fputc(ch2, stdout);
                break;
            case 152: /* Chinese Big 5 Level 1 */
            case 153: /* Chinese Big 5 Level 2 */
                if(!(last_enc == 152 || last_enc == 153))
                    fprintf(stdout, "\\CJKenc{Bg5}");
                ch1 = fgetc(stdin);
                ch2 = fgetc(stdin);@#

                temp = (ch1 - 0xA1) * (0xFF - 0xA1) + (ch2 - 0xA1);
                if(ch == 153)
                    temp += (0xFF - 0xA1 + 0x7F - 0x40) * (0xC9 - 0xA1);
                out1 = temp / (0xFF - 0xA1 + 0x7F - 0x40) + 0xA1;
                out2 = temp % (0xFF - 0xA1 + 0x7F - 0x40);
                out2 += (out2 < 0x3F) ? 0x40 : 0x62;@#

                fprintf(stdout, "%c%d.", out1, out2);
                break;@#
                
            case 147: /* Korean (KS) */
                if(last_enc != 147)
                    fprintf(stdout, "\\CJKenc{KS}");
                ch1 = fgetc(stdin);
                ch2 = fgetc(stdin);
                fputc(ch1, stdout);
                fputc(ch2, stdout);
                break;@#

            default:
                fputc(ch, stdout);
           }@#

        if(ch >= 0x80)
            last_enc = ch;
        ch = fgetc(stdin);
       }
    return 0;
   }
